% Spanish Stemmer 
% Dennis D. Perez

% This program implements a variant of the SNOWBALL spanish stemming algorithm
% This program was developed as a part of the PRONTO project at the Artificial Intelligence Center, 
% University of Georgia ( www.ai.uga.edu )

% Utility functions

accents(225,97).
accents(233,101).
accents(237,105).
accents(243,111).
accents(250,117). 
accents(252,117).

is_vowel(97).  % 'a'
is_vowel(101). % 'e'
is_vowel(105). % 'i'
is_vowel(111). % 'o'
is_vowel(117). % 'u'
is_vowel(225). % ''
is_vowel(233). % ''
is_vowel(237). % ''
is_vowel(243). % ''
is_vowel(250). % ''
is_vowel(252). % ''

is_allowed(Code) :- allowed_range(Min,Max), Min =< Code, Code =< Max, !.

allowed_range(32,  32). %whitespace
allowed_range(48,  57). %digits
allowed_range(65,  90).	%lowercase letters
allowed_range(97, 122). %uppercase letters
%allowed_range(129,130). % , 
%allowed_range(160,165). % 
allowed_range(252,252). %  
allowed_range(233,233). % 
allowed_range(225,225). % 
allowed_range(237,237). % 
allowed_range(243,243). % 
allowed_range(250,250). % 
allowed_range(241,241). % 
allowed_range(209,209). % 


% Reverse the list.

reverse_list(List, Result) :- reverse_list_aux(List, [], Result).
reverse_list_aux([H|T], Stack, Result) :- reverse_list_aux(T,[H|Stack],Result).
reverse_list_aux([],Result,Result).

% eliminate_accents(+OriginalWord, -NewWord)

eliminate_accents([H|T],[NH|NT]) :- accents(H,NH), !, eliminate_accents(T,NT).
eliminate_accents([H|T],[H|NT]) :- eliminate_accents(T,NT).

% step 0 : Check for attached pronouns

pronoun(A, B) :-
        (   '$append'("em", B, A)		%"me"
        ;   '$append'("es", B, A)		%"se"
        ;   '$append'("et", B, A)		%"te"
        ;   '$append'("olet", B, A)		%"telo"
        ;   '$append'("olem", B, A)		%"melo"
        ;   '$append'("solet", B, A)		%"telos"
        ;   '$append'("solem", B, A)		%"melos"
        ;   '$append'("alet", B, A)		%"tela"
        ;   '$append'("alem", B, A)		%"mela"
        ;   '$append'("salet", B, A)		%"telas"
        ;   '$append'("salem", B, A)		%"melas"
        ;   '$append'("ales", B, A)		%"sela"
        ;   '$append'("oles", B, A)		%"selo" 
        ;   '$append'("sales", B, A)		%"selas"	
        ;   '$append'("soles", B, A)		%"selos"
        ;   '$append'("al", B, A)		%"la"
        ;   '$append'("el", B, A)		%"le"
        ;   '$append'("ol", B, A)		%"lo"
        ;   '$append'("sal", B, A)		%"las"
        ;   '$append'("sel", B, A)		%"les"
        ;   '$append'("sol", B, A)		%"los"
        ;   '$append'("son", B, A)		%"nos"
        ).


prepronoun(A, A) :-
        (   '$append'("odni", B, A)		%"indo"
        ;   '$append'("odn", B, A)		%"ndo"
        ;   '$append'("r", B, A)		%"r"
        ;   '$append'("r", B, A)		%"r"
        ;   '$append'("r", B, A)		%"r"
        ;   '$append'("odnei", B, A)		%"iendo"
        ;   '$append'("odna", B, A)		%"ando"
        ;   '$append'("ra", B, A)		%"ar"
        ;   '$append'("re", B, A)		%"er"
        ;   '$append'("ri", B, A)		%"ir"
        ;   '$append'("odneyu", B, A)		%"uyendo" - 'u' could be in root area.
        ).

step0(WordInv,StemInv) :- pronoun(WordInv,WordTmp), prepronoun(WordTmp, StemInv), !.
step0(WordInv,WordInv).

% Step 1 : Remove Standard Suffixes
% standard_suffix(?Suffix,?ReplacingChars).   All parameters are inverted strings.

standard_suffix("azna",[]).				% "anza",""
standard_suffix("sazna",[]).				% "anzas",""
standard_suffix("oci",[]).				% "ico",""
standard_suffix("aci",[]).				% "ica",""
standard_suffix("soci",[]).				% "icos",""
standard_suffix("saci",[]).				% "icas",""
standard_suffix("omsi",[]).				% "ismo",""
standard_suffix("somsi",[]).				% "ismos",""
standard_suffix("elba",[]).				% "able",""
standard_suffix("selba",[]).				% "ables",""
standard_suffix("elbi",[]).				% "ible",""
standard_suffix("selbi",[]).				% "ibles",""
standard_suffix("atsi",[]).				% "ista",""
standard_suffix("satsi",[]).				% "istas",""
standard_suffix("oso",[]).				% "oso",""
standard_suffix("aso",[]).				% "osa",""
standard_suffix("soso",[]).				% "osos",""
standard_suffix("saso",[]).				% "osas",""
standard_suffix("otneima",[]).				% "amiento",""
standard_suffix("sotneima",[]).				% "amientos",""
standard_suffix("otneimi",[]).				% "imiento",""
standard_suffix("sotneimi",[]).				% "imientos",""

standard_suffix("arodaci",[]).				% "icadora",""
standard_suffix("rodaci",[]).				% "icador", ""
standard_suffix("nicaci",[]).				% "icacin",""
standard_suffix("sarodaci",[]).				% "icadoras",""
standard_suffix("serodaci",[]).				% "icadores",""
standard_suffix("senoicaci",[]).			% "icaciones",""
standard_suffix("etnaci",[]).				% "icante",""
standard_suffix("setnaci",[]).				% "icantes",""
standard_suffix("aicnaci",[]).				% "icancia",""
standard_suffix("saicnaci",[]).				% "icancias",""
standard_suffix("aroda",[]).				% "adora",""
standard_suffix("roda",[]).				% "ador", ""
standard_suffix("nica",[]).				% "acin",""
standard_suffix("saroda",[]).				% "adoras",""
standard_suffix("seroda",[]).				% "adores",""
standard_suffix("senoica",[]).				% "aciones",""
standard_suffix("etna",[]).				% "ante",""
standard_suffix("setna",[]).				% "antes",""
standard_suffix("aicna",[]).				% "ancia",""
standard_suffix("saicna",[]).				% "ancias",""
								
standard_suffix("agol","gol").				% "loga","log"
standard_suffix("sagol","gol").			% "logas","log"
												 				standard_suffix("nicu","u").				% "ucin","u"	 
standard_suffix("senicu","u").				% "ucines","u"

standard_suffix("acne","etne").			% "encia","ente"
standard_suffix("sacne","etne").			% "encias","ente"

standard_suffix("etnemavita",[]).			% "ativamente",""
standard_suffix("etnemavi",[]).				% "ivamente",""
standard_suffix("etnemaso",[]).				% "osamente",""
standard_suffix("etnemaci",[]).				% "icamente",""
standard_suffix("etnemada",[]).				% "adamente",""
standard_suffix("etnema",[]).				% "amente",""

standard_suffix("etnemetna",[]).			% "antemente",""
standard_suffix("etnemelba",[]).			% "ablemente",""
standard_suffix("etnemelbi",[]).			% "iblemente",""
standard_suffix("etnem",[]).				% "mente",""

standard_suffix("dadiliba",[]).				% "abilidad",""
standard_suffix("sedadiliba",[]).			% "abilidades",""
standard_suffix("dadici",[]).				% "icidad",""
standard_suffix("sedadici",[]).				% "icidades",""
standard_suffix("dadivi",[]).				% "ividad",""
standard_suffix("sedadivi",[]).				% "ividades",""
standard_suffix("dadi",[]).				% "idad",""
standard_suffix("sedadi",[]).				% "idades",""

standard_suffix("avita",[]).				% "ativa",""
standard_suffix("ovita",[]).				% "ativo",""
standard_suffix("savita",[]).				% "ativas",""
standard_suffix("sovita",[]).				% "ativos",""
standard_suffix("avi",[]).				% "iva",""
standard_suffix("ovi",[]).				% "ivo",""
standard_suffix("savi",[]).				% "ivas",""
standard_suffix("sovi",[]).				% "ivos",""


step1(WordInv,R2,StemInv) :- 	standard_suffix(Suffix, Replace), 
				'$append'(Suffix, NoSuffixR2, R2),  		% If exists in R2
				\+ NoSuffixR2 == [],			
				'$append'(Suffix, NoSuffix, WordInv),		% Remove it from RV !
				'$append'(Replace, NoSuffix, StemInv), !.  	% and replace it if needed.
step1(WordInv,_,WordInv).


% Step 2 : If no change was made on Step 1... additional suffixes

% First check for verb suffixes and delete them if preceded by 'u'.

verb_suffix("ay").		% "ya" 
verb_suffix("ey").		% "ye" 
verb_suffix("nay").		% "yan" 
verb_suffix("ney").		% "yen" 
verb_suffix("norey").		% "yeron" 
verb_suffix("odney").		% "yendo" 
verb_suffix("oy").		% "yo" 
verb_suffix("y").		% "y" 
verb_suffix("say").		% "yas" 
verb_suffix("sey").		% "yes" 
verb_suffix("siay").		% "yais" 
verb_suffix("somay").		% "yamos" 

% If no suffix was removed... 

verb_suffix_b("neu","g").			% "uen", "g"
verb_suffix_b("ne","").				% "en"
verb_suffix_b("seu","g").			% "ues", "g"
verb_suffix_b("se","").				% "es"
verb_suffix_b("siu","g").			% "uis", "g"
verb_suffix_b("si","").			% "is"
verb_suffix_b("someu","g").			% "uemos", "g"
verb_suffix_b("some","").			% "emos"
verb_suffix_b("nara","").			% "aran"
verb_suffix_b("sara","").			% "aras"
verb_suffix_b("nra","").			% "arn"
verb_suffix_b("sra","").			% "ars"
verb_suffix_b("siara","").			% "arais"
verb_suffix_b("ara","").			% "ara"
verb_suffix_b("sira","").			% "aris"
verb_suffix_b("somara","").			% "aramos"
verb_suffix_b("somera","").			% "aremos"
verb_suffix_b("ra","").			% "ar"
verb_suffix_b("ra","").			% "ar"
verb_suffix_b("nare","").			% "eran"
verb_suffix_b("sare","").			% "eras"
verb_suffix_b("nre","").			% "ern"
verb_suffix_b("sre","").			% "ers"
verb_suffix_b("siare","").			% "erais"
verb_suffix_b("are","").			% "era"
verb_suffix_b("sire","").			% "eris"
verb_suffix_b("somare","").			% "eramos"
verb_suffix_b("somere","").			% "eremos"
verb_suffix_b("re","").			% "er"
verb_suffix_b("re","").			% "er"
verb_suffix_b("nari","").			% "iran"
verb_suffix_b("sari","").			% "iras"
verb_suffix_b("nri","").			% "irn"
verb_suffix_b("sri","").			% "irs"
verb_suffix_b("siari","").			% "irais"
verb_suffix_b("ari","").			% "ira"
verb_suffix_b("siri","").			% "iris"
verb_suffix_b("somari","").			% "iramos"
verb_suffix_b("someri","").			% "iremos"
verb_suffix_b("ri","").			% "ir"
verb_suffix_b("ri","").			% "ir"
verb_suffix_b("aba","").			% "aba"
verb_suffix_b("ada","").			% "ada"
verb_suffix_b("adi","").			% "ida"
verb_suffix_b("a","").				% "a"
verb_suffix_b("ara","").			% "ara"
verb_suffix_b("arei","").			% "iera"
verb_suffix_b("da","").				% "ad"
verb_suffix_b("de","").				% "ed"
verb_suffix_b("di","").				% "id"
verb_suffix_b("esa","").			% "ase"
verb_suffix_b("esei","").			% "iese"
verb_suffix_b("etsa","").			% "aste"
verb_suffix_b("etsi","").			% "iste"
verb_suffix_b("na","").				% "an"
verb_suffix_b("naba","").			% "aban"
verb_suffix_b("na","").			% "an"
verb_suffix_b("nara","").			% "aran"
verb_suffix_b("narei","").			% "ieran"
verb_suffix_b("nesa","").			% "asen"
verb_suffix_b("nesei","").			% "iesen"
verb_suffix_b("nora","").			% "aron"
verb_suffix_b("norei","").			% "ieron"
verb_suffix_b("oda","").			% "ado"
verb_suffix_b("odi","").			% "ido"
verb_suffix_b("odna","").			% "ando"
verb_suffix_b("odnei","").			% "iendo"
verb_suffix_b("i","").				% "i"
verb_suffix_b("ra","").				% "ar"
verb_suffix_b("re","").				% "er"
verb_suffix_b("ri","").				% "ir"
verb_suffix_b("sa","").				% "as"
verb_suffix_b("saba","").			% "abas"
verb_suffix_b("sada","").			% "adas"
verb_suffix_b("sadi","").			% "idas"
verb_suffix_b("sa","").			% "as"
verb_suffix_b("sara","").			% "aras"
verb_suffix_b("sarei","").			% "ieras"
verb_suffix_b("sesa","").			% "ases"
verb_suffix_b("sesei","").			% "ieses"
verb_suffix_b("s","").				% "s"
verb_suffix_b("si","").			% "is"
verb_suffix_b("siaba","").			% "abais"
verb_suffix_b("sia","").			% "ais"
verb_suffix_b("siara","").			% "arais"
verb_suffix_b("siarei","").			% "ierais"
verb_suffix_b("siesa","").			% "aseis"
verb_suffix_b("siesei","").			% "ieseis"
verb_suffix_b("sietsa","").			% "asteis"
verb_suffix_b("sietsi","").			% "isteis"
verb_suffix_b("soda","").			% "ados"
verb_suffix_b("sodi","").			% "idos"
verb_suffix_b("soma","").			% "amos"
verb_suffix_b("somab","").			% "bamos"
verb_suffix_b("soma","").			% "amos"
verb_suffix_b("somi","").			% "imos"
verb_suffix_b("somar","").			% "ramos"
verb_suffix_b("somari","").			% "iramos"
verb_suffix_b("somesi","").			% "isemos"
verb_suffix_b("somes","").			% "semos"

step2(WordInv,StemInv) :- 	verb_suffix(Suffix),
				'$append'(Suffix, StemInv, WordInv),
				'$append'("u", _, StemInv), !.

step2(WordInv,StemInv) :- 	verb_suffix_b(Suffix, Precedence),
				'$append'(Suffix, StemInv, WordInv),
				'$append'(Precedence, NoPrecedence, StemInv),
				\+ NoPrecedence	== [], !.

step2(WordInv, WordInv).


% Step 3 & 4 : 

% 4
residual_suffix("oti","").		% ito
residual_suffix("ati","").		% ita
residual_suffix("oza","").		% azo
residual_suffix("aza","").		% aza
residual_suffix("nil","").		% lin
residual_suffix("anil","").		% lina
residual_suffix("ni","").		% in
residual_suffix("ani","").		% ina
residual_suffix("no","").		% on
residual_suffix("ano","").		% ona
residual_suffix("olliti","").		% itillo
residual_suffix("alliti","").		% itilla
residual_suffix("ollic","").		% cillo
residual_suffix("allic","").		% cilla
residual_suffix("olli","").		% illo
residual_suffix("alli","").		% illa
residual_suffix("ociti","").		% itico
residual_suffix("aciti","").		% itica
residual_suffix("oci","").		% ico
residual_suffix("aci","").		% ica
residual_suffix("eto","").		% ote
residual_suffix("ato","").		% ota
% 3
residual_suffix("so","").		% os
residual_suffix("a","").		% a
residual_suffix("o","").		% o
residual_suffix("","").		% 
residual_suffix("","").		% 
residual_suffix("","").		% 
residual_suffix("eu","g").		% g,ue
residual_suffix("u","g").		% g,u
residual_suffix("e","").		% e
residual_suffix("","").		% 


step3(WordInv, StemInv) :- 	residual_suffix(Suffix, Precedence),
				'$append'(Suffix, StemInv, WordInv),
				'$append'(Precedence, _, StemInv).

step3(WordInv, WordInv).

% removes the accents from the stem.

removeAccent([],[]).
removeAccent([H|T],[NH|NT]) :- accents(H,NH), removeAccent(T,NT).
removeAccent([H|T],[H|NT]) :- removeAccent(T,NT).


% Check for R1, R2 and RV.
% Note : RV excludes the first 3 letters of the word. Some precedence check may include the last letter in RV. 
% To simplify the algorithm, RV here will contain the last root letter.

getRV_R1_R2(WordInv,RV,R2,[A,B]) :- 	append(RV,[A,B],WordInv),
					getR2(WordInv,R2). 

% Finds R2: getR2(WordInv, R2).

getR2(WordInv, R2) :- getR2Aux(WordInv, R2, _).

getR2Aux([],[],0) :- !.
getR2Aux([H|T],NewR2,P) :- getR2Aux(T,R2,S),!, ( S == 0, '$append'([], R2, NewR2),
							( is_vowel(H), P is 1
								;
				  			  P is 0 )
						;
						P is 1 , '$append'([H], R2, NewR2)).


% The stemming algorithm itself.

stemWordInv(WordInv, StemInvFinal) :- 	getRV_R1_R2(WordInv,RV,R2,NonRV),
					step0(RV, StemInv0),
					step1(StemInv0, R2, StemInv1),
					( \+ StemInv0 == StemInv1, '$append'([],StemInv1,StemInv2)  % If nothing was removed, 
					;	
					  step2(StemInv0,StemInv2)	),			% do step 2.
					step3(StemInv2,StemInv3),					
					append(StemInv3,NonRV,StemInv),
					removeAccent(StemInv,StemInvFinal).

stemWordInv(WordInv, WordInvFinal) :- removeAccent(WordInv, WordInvFinal).


% OPTION 1 : Stem a single word.

stemWord(Word, Stem) :- reverse_list(Word, WordInv), stemWordInv(WordInv,StemInv), reverse_list(StemInv,Stem).


% OPTION 2 : Read the file we're going to stem, and stem each word.

% First build a list with all the characters.
read_line([]) :- at_end_of_stream, !.
read_line([]) :- peek_code(Char), Char == 10, get_code(_), !.
%read_line([]) :- peek_code(Char), Char == 13, get_code(_), !.
read_line([32|List]) :- peek_code(Char), \+ is_allowed(Char), get_code(_), read_line(List), !. 
read_line([Char|List]) :- get_code(Char), read_line(List), !.

% Then tokenize it so each word could be stemmed 
% NOTE : The tokens are inverse, just as we process them with the stemmer !!!

tokenize_stem_list(List) :- tokenize_list_aux(List, []).

tokenize_list_aux([], TokensInv) :- stemWordInv(TokensInv, StemInv), printStemInv(StemInv), !. 	% The last word.
tokenize_list_aux([Char|List], TokensInv) :- 	Char == 32,
						stemWordInv(TokensInv, StemInv), printStemInv(StemInv), !, 
					    	tokenize_list_aux(List, []).
tokenize_list_aux([Char|List], Tokens) :- tokenize_list_aux(List, [Char|Tokens]).


processLine :- read_line(List), tokenize_stem_list(List), nl.

printStemInv([]) :- nl.
printStemInv([H|T]) :- printStemInv(T), char_code(Hchr,H), write(Hchr).

% Applies the stemming algorithm to a file.

stemFile(InputFileChr, OutputFileChr) :- atom_chars(InputFile, InputFileChr), 
					see(InputFile), 
					atom_chars(OutputFile, OutputFileChr),
					tell(OutputFile), 
					repeat,
					processLine, 
					at_end_of_stream,
					seen, 					
					told,
					write('Stem File Generated.'), 
					nl.





